@@ -13,7 +13,7 @@
 ***********************************************************************/
 
 static char const RCSID[] =
-"$Id: utils.c,v 1.18 2003/02/14 18:28:29 dfs Exp $";
+"$Id: utils.c,v 1.36 2004/03/26 19:33:33 dfs Exp $";
 
 #define _BSD_SOURCE 1
 
@@ -26,6 +26,7 @@ static char const RCSID[] =
 #include "config.h"
 #include "mimedefang.h"
 
+#include <stdio.h>
 #include <ctype.h>
 #include <syslog.h>
 #include <stdlib.h>
@@ -36,11 +37,30 @@ static char const RCSID[] =
 #include <unistd.h>
 #include <errno.h>
 #include <stdarg.h>
+#include <netinet/in.h>
+
+#ifdef HAVE_STDINT_H
+#include <stdint.h>
+#endif
+
+#ifndef HAVE_UINT32_T
+/* On these machines, punt to unsigned int */
+typedef unsigned int uint32_t;
+#endif
 
 #ifndef AF_LOCAL
 #define AF_LOCAL AF_UNIX
 #endif
 
+#ifndef SUN_LEN
+#define SUN_LEN(ptr)  ((size_t) (((struct sockaddr_un *) 0)->sun_path) \
+       + strlen ((ptr)->sun_path))
+#endif
+
+#ifndef INADDR_LOOPBACK
+#define INADDR_LOOPBACK 0x7f000001
+#endif
+
 /**********************************************************************
 * %FUNCTION: split_on_space
 * %ARGUMENTS:
@@ -163,10 +183,11 @@ chomp(char *str)
 /**********************************************************************
 * %FUNCTION: safeWriteHeader
 * %ARGUMENTS:
-*  fp -- file to write to
+*  fd -- file to write to
 *  str -- a string value
 * %RETURNS:
 *  0 if header seems OK; 1 if suspicious character found.
+*  Note that "str" MAY BE modified!
 * %DESCRIPTION:
 *  Writes "str" to FP with the following changes:
 *    CRLF -> written as-is
@@ -175,22 +196,23 @@ chomp(char *str)
 *    LF   -> written as CRLF
 ***********************************************************************/
 int
-safeWriteHeader(FILE *fp,
-		unsigned char const *str)
+safeWriteHeader(int fd,
+		unsigned char *str)
 {
     int suspicious = 0;
+    unsigned char *origStr = str;
+
     for(; *str; str++) {
-	/* Do not write \r to header file */
+	/* Do not write \r to header file -- convert to space*/
 	if (*str == '\r') {
 	    if (*(str+1) != '\n') {
 		suspicious = 1;
+		*str = ' ';
 	    }
-	    continue;
 	}
-
-	putc(*str, fp);
     }
 
+    writestr(fd, origStr);
     return suspicious;
 }
 
@@ -214,7 +236,7 @@ MXCommand(char const *sockname,
 {
     int fd;
     struct sockaddr_un addr;
-    int nleft, nwritten, nread;
+    int nread;
     int n;
 
     fd = socket(AF_LOCAL, SOCK_STREAM, 0);
@@ -233,18 +255,7 @@ MXCommand(char const *sockname,
 	return -1;
     }
 
-    nwritten = 0;
-    nleft = strlen(cmd);
-    n = 0;
-    while(nleft) {
-	n = write(fd, cmd+nwritten, nleft);
-	if (n < 0) {
-	    if (errno == EINTR) continue;
-	    break;
-	}
-	nwritten += n;
-	nleft -= n;
-    }
+    n = writestr(fd, cmd);
     if (n < 0) {
 	syslog(LOG_ERR, "MXCommand: write: %m: Is multiplexor running?");
 	close(fd);
@@ -252,18 +263,8 @@ MXCommand(char const *sockname,
     }
 
     /* Now read the answer */
-    nread = 0;
-    nleft = len;
-    while(nleft) {
-	n = read(fd, buf+nread, nleft);
-	if (n <= 0) {
-	    if (n < 0 && errno == EINTR) continue;
-	    break;
-	}
-	nread += n;
-	nleft -= n;
-    }
-    if (n < 0) {
+    nread = readn(fd, buf, len-1);
+    if (nread < 0) {
 	syslog(LOG_ERR, "MXCommand: read: %m: Is multiplexor running?");
 	close(fd);
 	return -1;
@@ -328,7 +329,7 @@ MXScanDir(char const *sockname,
 * %FUNCTION: write_percent_encoded
 * %ARGUMENTS:
 *  str -- a string
-*  fp -- file to write to
+*  fd -- file to write to
 * %RETURNS:
 *  Nothing
 * %DESCRIPTION:
@@ -339,14 +340,36 @@ MXScanDir(char const *sockname,
 ***********************************************************************/
 void
 write_percent_encoded(unsigned char const *str,
-		      FILE *fp)
+		      int fd)
 {
+    unsigned char buf[2048];
     int c;
-    while ((c = *str++) != 0) {
-	if (c <= 32 || c > 126 || c == '%' || c == '\\' || c == '\'' || c == '"') {
-	    fprintf(fp, "%%%02X", (unsigned int) c);
-	} else {
-	    fputc(c, fp);
+    unsigned char *s = buf;
+    if (strlen(str) < 640) {
+	/* Fast -- create in memory buffer, then write out all at once */
+	while ((c = *str++) != 0) {
+	    if (c <= 32 || c > 126 || c == '%' ||
+		c == '\\' || c == '\'' || c == '"') {
+		sprintf(s, "%%%02X", (unsigned int) c);
+		s += 3;
+	    } else {
+		*s++ = c;
+	    }
+	}
+	/* Null termination -- not strictly required because of writen */
+	*s = 0;
+	writen(fd, buf, s-buf);
+    } else {
+	/* Slow -- string too long; use individual writes */
+	while ((c = *str++) != 0) {
+	    if (c <= 32 || c > 126 || c == '%' ||
+		c == '\\' || c == '\'' || c == '"') {
+		sprintf(s, "%%%02X", (unsigned int) c);
+		writen(fd, s, 3);
+	    } else {
+		*s = c;
+		writen(fd, s, 1);
+	    }
 	}
     }
 }
@@ -489,6 +512,7 @@ percent_encode_command(char *out, int outlen, ...)
 *  name -- relay name
 *  helo -- argument to "HELO/EHLO" (may be NULL)
 *  dir -- MIMEDefang working directory
+*  qid -- Sendmail queue identifier
 * %RETURNS:
 *  1 if it's OK to accept connections from this host; 0 if not, -1 if error.
 *  If connection is rejected, error message *may* be set.
@@ -499,7 +523,8 @@ MXRelayOK(char const *sockname,
 	  char const *ip,
 	  char const *name,
 	  char const *helo,
-	  char const *dir)
+	  char const *dir,
+	  char const *qid)
 {
     char cmd[SMALLBUF];
     char ans[SMALLBUF];
@@ -514,14 +539,15 @@ MXRelayOK(char const *sockname,
 	name = ip;
     }
     if (!helo) {
-	helo = "";
+	helo = "UNKNOWN";
     }
     percent_encode_command(cmd, sizeof(cmd), "relayok", ip, name, helo,
-			   dir, NULL);
+			   dir, qid, NULL);
     if (MXCommand(sockname, cmd, ans, SMALLBUF-1) < 0) return -1;
     if (!strcmp(ans, "ok -1\n")) return -1;
     if (!strcmp(ans, "ok 1\n")) return 1;
     if (!strcmp(ans, "ok 2\n")) return 2;
+    if (!strcmp(ans, "ok 3\n")) return 3;
     if (!strcmp(ans, "ok 0\n")) return 0;
 
     chomp(ans);
@@ -554,6 +580,7 @@ MXRelayOK(char const *sockname,
 *  name -- sending relay's host name
 *  helo -- argument to "HELO/EHLO" (may be NULL)
 *  dir -- MIMEDefang working directory
+*  qid -- Sendmail queue identifier
 * %RETURNS:
 *  1 if it's OK to accept messages from this sender; 0 if not, -1 if error or
 *  we should tempfail.
@@ -566,7 +593,8 @@ MXSenderOK(char const *sockname,
 	   char const *ip,
 	   char const *name,
 	   char const *helo,
-	   char const *dir)
+	   char const *dir,
+	   char const *qid)
 {
     char cmd[SMALLBUF];
     char ans[SMALLBUF];
@@ -585,14 +613,15 @@ MXSenderOK(char const *sockname,
 	name = ip;
     }
     if (!helo) {
-	helo = "";
+	helo = "UNKNOWN";
     }
     percent_encode_command(cmd, sizeof(cmd), "senderok", sender, ip, name,
-			   helo, dir, NULL);
+			   helo, dir, qid, NULL);
     if (MXCommand(sockname, cmd, ans, SMALLBUF-1) < 0) return -1;
     if (!strcmp(ans, "ok -1\n")) return -1;
     if (!strcmp(ans, "ok 1\n")) return 1;
     if (!strcmp(ans, "ok 2\n")) return 2;
+    if (!strcmp(ans, "ok 3\n")) return 3;
     if (!strcmp(ans, "ok 0\n")) return 0;
 
     chomp(ans);
@@ -626,6 +655,10 @@ MXSenderOK(char const *sockname,
 *  firstRecip -- first recipient of the message
 *  helo -- argument to "HELO/EHLO" (may be NULL)
 *  dir -- MIMEDefang working directory
+*  qid -- Sendmail queue identifier
+*  rcpt_mailer -- the "mailer" part of the triple for RCPT TO address
+*  rcpt_host -- the "host" part of the triple for RCPT TO address
+*  rcpt_addr -- the "addr" part of the triple for RCPT TO address
 * %RETURNS:
 *  1 if it's OK to accept messages to this recipient; 0 if not, -1 if error.
 *  If recipient is rejected, error message *may* be set.
@@ -639,7 +672,12 @@ MXRecipientOK(char const *sockname,
 	      char const *name,
 	      char const *firstRecip,
 	      char const *helo,
-	      char const *dir)
+	      char const *dir,
+	      char const *qid,
+	      char const *rcpt_mailer,
+	      char const *rcpt_host,
+	      char const *rcpt_addr)
+
 {
     char cmd[SMALLBUF];
     char ans[SMALLBUF];
@@ -666,16 +704,18 @@ MXRecipientOK(char const *sockname,
 	firstRecip = "UNKNOWN";
     }
     if (!helo) {
-	helo = "";
+	helo = "UNKNOWN";
     }
 
     percent_encode_command(cmd, sizeof(cmd),
 			   "recipok", recipient, sender, ip, name, firstRecip,
-			   helo, dir, NULL);
+			   helo, dir, qid, rcpt_mailer, rcpt_host, rcpt_addr,
+			   NULL);
     if (MXCommand(sockname, cmd, ans, SMALLBUF-1) < 0) return -1;
     if (!strcmp(ans, "ok -1\n")) return -1;
     if (!strcmp(ans, "ok 1\n")) return 1;
     if (!strcmp(ans, "ok 2\n")) return 2;
+    if (!strcmp(ans, "ok 3\n")) return 3;
     if (!strcmp(ans, "ok 0\n")) return 0;
 
     chomp(ans);
@@ -700,7 +740,7 @@ MXRecipientOK(char const *sockname,
 /**********************************************************************
 * %FUNCTION: write_mx_command
 * %ARGUMENTS:
-*  fp -- file to write to
+*  fd -- file to write to
 *  cmd -- the command to write.  A single character.
 *  buf -- command arguments
 * %RETURNS:
@@ -709,14 +749,361 @@ MXRecipientOK(char const *sockname,
 *  Writes a command for communicating with filter.
 ***********************************************************************/
 void
-write_mx_command(FILE *fp,
+write_mx_command(int fd,
 		 char cmd,
 		 unsigned char const *buf)
 {
-    if (!fp) return;
-    fprintf(fp, "%c", cmd);
+    unsigned char newline = '\n';
+    if (fd < 0) return;
+    writen(fd, (unsigned char *) &cmd, 1);
     if (buf) {
-	write_percent_encoded(buf, fp);
+	write_percent_encoded(buf, fd);
     }
-    fprintf(fp, "\n");
+    writen(fd, &newline, 1);
+}
+
+/**********************************************************************
+* %FUNCTION: writen
+* %ARGUMENTS:
+*  fd -- file to write to
+*  buf -- buffer to write
+*  len -- length to write
+* %RETURNS:
+*  Number of bytes written, or -1 on error
+* %DESCRIPTION:
+*  Writes exactly "len" bytes from "buf" to file descriptor fd
+***********************************************************************/
+int
+writen(int fd,
+       unsigned char const *buf,
+       size_t len)
+{
+    int r;
+    int nleft = len;
+    while(nleft) {
+	r = write(fd, buf, nleft);
+	if (r > 0) {
+	    nleft -= r;
+	    buf += r;
+	    continue;
+	}
+	if (r == 0) {
+	    /* Shouldn't happen! */
+	    errno = EIO;
+	    return -1;
+	}
+	if (r < 0) {
+	    if (errno == EINTR || errno == EAGAIN) {
+		continue;
+	    }
+	}
+	return r;
+    }
+    return len;
+}
+
+/**********************************************************************
+* %FUNCTION: writestr
+* %ARGUMENTS:
+*  fd -- file to write to
+*  buf -- null-terminated string to write
+* %RETURNS:
+*  Number of bytes written, or -1 on error
+* %DESCRIPTION:
+*  Writes the string in "buf" to fd.
+***********************************************************************/
+int
+writestr(int fd,
+	 unsigned char const *buf)
+{
+    return writen(fd, buf, strlen(buf));
+}
+
+/**********************************************************************
+* %FUNCTION: readn
+* %ARGUMENTS:
+*  fd -- file descriptor to read from
+*  buf -- buffer to read into
+*  count -- number of bytes to read
+* %RETURNS:
+*  The number of bytes actually read, or -1 on error
+* %DESCRIPTION:
+*  Attempts to read exactly "count" bytes from a descriptor.
+***********************************************************************/
+int
+readn(int fd, void *buf, size_t count)
+{
+    size_t num_read = 0;
+    unsigned char *c = (unsigned char *) buf;
+    while (count) {
+	int n = read(fd, c, count);
+	if (n == 0) { /* EOF */
+	    return num_read;
+	}
+	if (n < 0) { /* Error */
+	    if (errno == EINTR || errno == EAGAIN) {
+		continue;
+	    }
+	    return n;
+	}
+	num_read += n;
+	count -= n;
+	c += n;
+    }
+    return num_read;
+}
+
+/**********************************************************************
+* %FUNCTION: closefd
+* %ARGUMENTS:
+*  fd -- file to close
+* %RETURNS:
+*  Whatever close(2) returns
+* %DESCRIPTION:
+*  Closes fd, handling EINTR
+***********************************************************************/
+int
+closefd(int fd)
+{
+    int r;
+    while(1) {
+	r = close(fd);
+	if (r >= 0) return r;
+	if (errno != EINTR) return r;
+    }
+}
+
+/**********************************************************************
+* %FUNCTION: validate_smtp_code
+* %ARGUMENTS:
+*  code -- an SMTP code (eg 451)
+*  first -- what the first char must be
+* %RETURNS:
+*  1 if it's a valid code; 0 otherwise.  A valid code consists
+*  of three decimal digits only.  The first digit must match "first".
+***********************************************************************/
+int
+validate_smtp_code(char const *code,
+		   char const *first)
+{
+    if (!code) return 0;
+    if (*code != *first) return 0;
+
+    if (!isdigit(*(code+1))) return 0;
+    if (!isdigit(*(code+2))) return 0;
+    if (*(code+3)) return 0;
+
+    return 1;
+}
+
+/**********************************************************************
+* %FUNCTION: validate_smtp_dsn
+* %ARGUMENTS:
+*  dsn -- an SMTP dsn reply (eg 4.7.1)
+*  first -- what the first char must be
+* %RETURNS:
+*  1 if it's a valid dsn; 0 otherwise.  A valid DSN consits of three
+*  numerical fields separated by periods.  The first field must be a
+*  single digit that matches "first".  The second and
+*  third fields can be 1-3 digits long each.
+***********************************************************************/
+int
+validate_smtp_dsn(char const *dsn,
+		  char const *first)
+{
+    char const *s;
+    int count;
+
+    if (!dsn) return 0;
+    if (*dsn != *first) return 0;
+
+    if (*(dsn+1) != '.') return 0;
+
+    s = dsn+2;
+    count = 0;
+    while (isdigit(*s) && count < 4) {
+	count++;
+	s++;
+    }
+
+    if (count == 0 || count > 3) return 0;
+    if (*s != '.') return 0;
+
+    s++;
+    count = 0;
+    while (isdigit(*s) && count < 4) {
+	count++;
+	s++;
+    }
+    if (count == 0 || count > 3) return 0;
+    if (*s) return 0;
+
+    return 1;
+}
+
+/**********************************************************************
+* %FUNCTION: make_listening_socket
+* %ARGUMENTS:
+*  str -- a string of the form:
+*                   /path/to/sock   (assumed to be unix:/path/to/sock)
+*                   unix:/path/to/sock
+*                   local:/path/to/sock
+*                   inet:port  (host defaults to LOOPBACK)
+*  backlog -- listen backlog.  If -1, we use default of 5
+* %RETURNS:
+*  A listening UNIX-domain or TCP socket
+* %DESCRIPTION:
+*  Utility function for opening a listening socket.
+***********************************************************************/
+int
+make_listening_socket(char const *str, int backlog)
+{
+    char const *path;
+    int sock;
+    struct sockaddr_in addr_in;
+    struct sockaddr_un addr_un;
+    int opt;
+    int port;
+
+    if (backlog <= 0) backlog = 5;
+
+    if (!strncmp(str, "inet:", 5)) {
+	path = str+5;
+	if (sscanf(path, "%d", &port) != 1 ||
+	    port < 1 ||
+	    port > 65535) {
+	    syslog(LOG_ERR, "make_listening_socket: Invalid port %s", path);
+	    errno = EINVAL;
+	    return -1;
+	}
+
+	sock = socket(PF_INET, SOCK_STREAM, 0);
+	if (sock < 0) {
+	    syslog(LOG_ERR, "make_listening_socket: socket: %m");
+	    return -1;
+	}
+
+	opt = 1;
+	/* Reuse port */
+	if (setsockopt(sock, SOL_SOCKET, SO_REUSEADDR, &opt, sizeof(opt)) < 0) {
+	    syslog(LOG_ERR, "make_listening_socket: setsockopt: %m");
+	    close(sock);
+	    return -1;
+	}
+
+	addr_in.sin_family = AF_INET;
+	addr_in.sin_port = htons(port);
+	addr_in.sin_addr.s_addr = (uint32_t) htonl(INADDR_LOOPBACK);
+	if (bind(sock, (struct sockaddr *) &addr_in, sizeof(addr_in)) < 0) {
+	    syslog(LOG_ERR, "make_listening_socket: bind: %m");
+	    close(sock);
+	    return -1;
+	}
+    } else {
+	/* Assume unix-domain socket */
+	path = str;
+	if (!strncmp(str, "unix:", 5)) path = str+5;
+	else if (!strncmp(str, "local:", 6)) path = str+6;
+	(void) remove(path);
+	sock = socket(AF_LOCAL, SOCK_STREAM, 0);
+	if (sock < 0) {
+	    syslog(LOG_ERR, "make_listening_socket: socket: %m");
+	    return -1;
+	}
+
+	memset(&addr_un, 0, sizeof(addr_un));
+	addr_un.sun_family = AF_LOCAL;
+	strncpy(addr_un.sun_path, path, sizeof(addr_un.sun_path)-1);
+
+	if (bind(sock, (struct sockaddr *) &addr_un, SUN_LEN(&addr_un)) < 0) {
+	    syslog(LOG_ERR, "make_listening_socket: bind: %m");
+	    close(sock);
+	    return -1;
+	}
+    }
+
+    if (listen(sock, backlog) < 0) {
+	/* Maybe backlog is too high... try again */
+	if (backlog > 5) {
+	    if (listen(sock, 5) < 0) {
+		syslog(LOG_ERR, "make_listening_socket: listen: %m");
+		close(sock);
+		return -1;
+	    }
+	}
+	syslog(LOG_ERR, "make_listening_socket: listen: %m");
+	close(sock);
+	return -1;
+    }
+    return sock;
+}
+
+/**********************************************************************
+* %FUNCTION: connect_to_socket
+* %ARGUMENTS:
+*  str -- a string of the form:
+*                   /path/to/sock   (assumed to be unix:/path/to/sock)
+*                   unix:/path/to/sock
+*                   local:/path/to/sock
+*                   inet:port  (host defaults to LOOPBACK)
+* %RETURNS:
+*  Connects on specified UNIX-domain or TCP socket.
+* %DESCRIPTION:
+*  Utility function for opening a connected socket socket.
+***********************************************************************/
+int
+connect_to_socket(char const *str)
+{
+    char const *path;
+    int sock;
+    struct sockaddr_in addr_in;
+    struct sockaddr_un addr_un;
+    int port;
+
+    if (!strncmp(str, "inet:", 5)) {
+	path = str+5;
+	if (sscanf(path, "%d", &port) != 1 ||
+	    port < 1 ||
+	    port > 65535) {
+	    syslog(LOG_ERR, "connect_to_socket: Invalid port %s", path);
+	    errno = EINVAL;
+	    return -1;
+	}
+
+	sock = socket(PF_INET, SOCK_STREAM, 0);
+	if (sock < 0) {
+	    syslog(LOG_ERR, "connect_to_socket: socket: %m");
+	    return -1;
+	}
+
+	addr_in.sin_family = AF_INET;
+	addr_in.sin_port = htons(port);
+	addr_in.sin_addr.s_addr = (uint32_t) htonl(INADDR_LOOPBACK);
+	if (connect(sock, (struct sockaddr *) &addr_in, sizeof(addr_in)) < 0) {
+	    syslog(LOG_ERR, "connect_to_socket: connect: %m");
+	    close(sock);
+	    return -1;
+	}
+    } else {
+	/* Assume unix-domain socket */
+	path = str;
+	if (!strncmp(str, "unix:", 5)) path = str+5;
+	else if (!strncmp(str, "local:", 6)) path = str+6;
+	sock = socket(AF_LOCAL, SOCK_STREAM, 0);
+	if (sock < 0) {
+	    syslog(LOG_ERR, "connect_to_socket: socket: %m");
+	    return -1;
+	}
+
+	memset(&addr_un, 0, sizeof(addr_un));
+	addr_un.sun_family = AF_LOCAL;
+	strncpy(addr_un.sun_path, path, sizeof(addr_un.sun_path)-1);
+
+	if (connect(sock, (struct sockaddr *) &addr_un, SUN_LEN(&addr_un)) < 0) {
+	    syslog(LOG_ERR, "connect_to_socket: connect: %m");
+	    close(sock);
+	    return -1;
+	}
+    }
+    return sock;
 }
